#pragma once


#include<dune/pdelab/common/quadraturerules.hh>
#include<dune/pdelab/finiteelement/localbasiscache.hh>
//#include<dune/pdelab/localoperator/idefault.hh>
#include<dune/pdelab/localoperator/defaultimp.hh>
#include<dune/pdelab/localoperator/pattern.hh>
#include<dune/pdelab/localoperator/flags.hh>
#include<dune/pdelab/localoperator/variablefactories.hh>
#include<dune/pdelab/gridfunctionspace/gridfunctionspace.hh>

template<typename CHGFS>
class PF_3P_DIFF_ESTIMATOR :
  public Dune::PDELab::LocalOperatorDefaultFlags
{
private:
  typedef typename CHGFS::template Child<0>::Type FEM;
  typedef typename FEM::Traits::FiniteElementType::Traits::LocalBasisType LocalBasis;
  Dune::PDELab::LocalBasisCache<LocalBasis> cache;
  enum {dim=LocalBasis::Traits::dimDomain};

public:
  enum {doPatternVolume=false};
  enum {doAlphaVolume=true};

  //constructor
  PF_3P_DIFF_ESTIMATOR()
  {
  }


  template<typename EG, typename LFSU, typename X, typename LFSV, typename R>
  void alpha_volume (const EG& eg, const LFSU& lfsu, const X& x, const LFSV& lfsv, R& r) const
  {
    using namespace Dune::Indices;
    const auto& pf1space = child(lfsu,_0);
    const auto& pf2space = child(lfsu,_2);
    typedef decltype(makeZeroBasisFieldValue(pf1space)) RF;

    auto geo = eg.geometry();
    const auto& rElement = Dune::ReferenceElements<double, dim>::general(geo.type());
    RF u1max=0.0;
    RF u1min=0.0;
    RF u2max=0.0;
    RF u2min=0.0;
    RF u3max=0.0;
    RF u3min=0.0;
  //  Dune::FieldVector<RF, dim> xg = eg.geometry().global( rElement.position(0, dim) );
    for (int i=0; i<geo.corners(); i++)
    {
      RF u1=0.0;
      RF u2=0.0;
      auto& phihat = cache.evaluateFunction(rElement.position(i, dim),pf1space.finiteElement().localBasis());
      for (size_t j=0; j<pf1space.size(); j++)
      {
        u1 += x(pf1space,j)*phihat[j];
        u2 += x(pf2space,j)*phihat[j];
      }
      RF u3=1-u1-u2;
      if(i==0)
      {
        u1max = u1;
        u1min = u1;
        u2max = u2;
        u2min = u2;
        u3max = u3;
        u3min = u3;
      }
      else
      {
        if(u1>u1max) u1max = u1;
        if(u1<u1min) u1min = u1;
        if(u2>u2max) u2max = u2;
        if(u2<u2min) u2min = u2;
        if(u3>u3max) u3max = u3;
        if(u3<u3min) u3min = u3;
      }
      //xg = eg.geometry().global( rElement.position(i, dim) );
    }
    r.accumulate(lfsv,0,0.5*((u1max-u1min)+(u2max-u2min)+(u3max-u3min)));//(xg[0]>0.25&&xg[0]<0.75)? (umax-umin):0 );
  }

};
